#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <libgen.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>
#include <mntent.h>  // getmntent_r()
#include "hopman.h"
#include "watch.h"

static FILE *pf = NULL;
static int debug = -1;

// Usage: strncpy_t(str1, sizeof(str1), str2, strlen(str2));
// Copy string "in" with at most "insz" chars to buffer "out", which
// is "outsz" bytes long. The output is always 0-terminated. Unlike
// strncpy(), strncpy_t() does not zero fill remaining space in the
// output buffer:
static
char* strncpy_t(char* out, size_t outsz, const char* in, size_t insz)
{
   assert(outsz > 0);
   while(--outsz > 0 && insz > 0 && *in) { *out++ = *in++; insz--; }
   *out = 0;
   return out;
} 	

/* On 1st call open mountinfo, on all calls return the file descriptor */
int mountinit(void)
{
  if(debug == -1)
    {
      debug = config_bool(Debug);
      if(debug == -1) debug = 0;
    }
  
  if(!pf)
    {
      pf=fopen("/proc/mounts", "r");
      if(!pf)
	{
	  switch(errno)
	    {
	    case ENOMEM:
	      fprintf(stderr, _("%s error opening /proc/mounts: %s\n"),
		      progname, strerror(errno));
	      break;
	    default:
	      fprintf( stderr, _("%s: %s: /proc/mounts.\n"),
		       progname, strerror(errno) );
	      exit(EXIT_FAILURE);
	    }
	}
    }
  return fileno(pf);
}

void mountpoints(void)
{
  /* read /proc/mounts and decode */
  /* it is impossible to stat the size of this file, nor to report
     its size by invoking lseek(fd, 0, SEEK_END). The first method
     gives 0 and the second -1. We resort to reading it line by line
     and the easiest for this is stdio. */

  partlist_t P;
  unsigned anychange;

  P = partition_get_list();
  if(!P.np) return;

  /*--------------- open on first call, rewind on other calls ---------------*/
  if(!pf) mountinit();
  else fseek(pf, 0, SEEK_SET);

  /*----- Dynamically allocate two arrays in the stack and update data ------*/
  {
    unsigned index;
    unsigned nmounts[P.np];
    char changed[P.np];	
    char mntbuf[PM_MNT_LEN+1];
    struct mntent ment_buf;
    struct mntent* ment_ptr = NULL;

    /* reset mount counters and modification flags */
    for(index=0; index<P.np; index++) changed[index] = nmounts[index] = 0;
    anychange = 0;
    
    /* decode every line of /proc/mounts */
    while( ment_ptr = getmntent_r( pf, &ment_buf, mntbuf, PM_MNT_LEN+1 ) )
      {
	char *partitionname = basename(ment_ptr->mnt_fsname);
	
	/* find if the partition is in our list */
	index = partition_by_name(partitionname);
	if( index < P.np )
	  {
	    /* yes, here's the partition */
	    if(!strcmp(ment_ptr->mnt_dir, "/run/live/medium") || !strcmp(ment_ptr->mnt_dir, "/lib/live/mount/medium"))
	      {
	         ui_delete(P.p[index]);
	      }   
        else
          {   
	         nmounts[index]++; /* increment the number of partition's mounts */
	         if( nmounts[index]==1 && strcmp(ment_ptr->mnt_dir, P.p[index]->mnt) )
	         {
		       /* It is the 1st mountpoint of the partition and it changed */
		       fprintf(stderr, "1st mountpoint changed on partition #%d\n", index);
		       strncpy_t(P.p[index]->fstype, PM_FSTYPE_LEN, ment_ptr->mnt_type, strlen(ment_ptr->mnt_type));
		       strncpy_t(P.p[index]->mnt, PM_MNT_LEN, ment_ptr->mnt_dir, strlen(ment_ptr->mnt_dir));
		       //strncpy( P.p[index]->fstype, fstp, PM_FSTYPE_LEN);
		       //strncpy(P.p[index]->mnt, mtpt, PM_MNT_LEN);
		       changed[index] = 1;
		       anychange = 1;
	         }
	      }
      }
      }

    /* Tag all partitions for which the number of mountpoints has changed and
       nullify the mountpoint string when the number of mountpoints is 0 */
    for(index=0; index<P.np; index++)
      {
	if( P.p[index]->nmounts != nmounts[index] )
	  {
	    P.p[index]->nmounts = nmounts[index];
	    changed[index] = 1;
	    anychange = 1;
	    if( nmounts[index]==0 ) P.p[index]->mnt[0] = '\0';
	  }
      }

    /* invoke ui_update for all changed partitions */
    for( index=0; index<P.np; index ++ )
      {
	if( changed[index] ) ui_update( P.p[index] );
      }
    /* all done */
  }

  if(debug && anychange) partition_print();

  return;
}
