/* See LICENSE file for copyright and license details. */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>   // readlink()
#include <assert.h>
#include <stdbool.h>
#include <sys/wait.h>
#include "hopman.h"
#include "watch.h"

/*=============================== WARNING ===================================*/
/* Beware: realloc() is used to increase/decrease the size of the memory     */
/* area used to store partition data, but realloc() may move this memory     */
/* area. Therefore pointers to partition's info are short-lived and cannot   */
/* be retained by the GUI, eg, for use by the callbacks.                     */
/*======================== THIS IS STILL TO BE DONE==========================*/
static partlist_t P = {0, NULL};
static int debug = -1;

// Usage: strncpy_t(str1, sizeof(str1), str2, strlen(str2));
// Copy string "in" with at most "insz" chars to buffer "out", which
// is "outsz" bytes long. The output is always 0-terminated. Unlike
// strncpy(), strncpy_t() does not zero fill remaining space in the
// output buffer:
static
char* strncpy_t(char* out, size_t outsz, const char* in, size_t insz)
{
   assert(outsz > 0);
   while(--outsz > 0 && insz > 0 && *in) { *out++ = *in++; insz--; }
   *out = 0;
   return out;
}

/*------------------------------- Creator -----------------------------------*/
void partition_new( ino_t inode, const char *name )
{
  char *buf = NULL;
  
  if(debug==-1)
    {
      debug = config_bool(Debug);
      if(debug == -1) debug = 0;
    }
  /* reallocate memory for the new array of partinfo_t pointers */
  P.p = realloc( P.p, sizeof(partinfo_t *) * (P.np+1) );

  if( P.np && !P.p )
    {
      fprintf( stderr, _("%s Cannot store partition data: %s\n"),
	       progname, strerror(errno) );
      exit( EXIT_FAILURE );
    }
  
  /* Allocate memory for this new partition */
  P.p[P.np] = malloc( sizeof(partinfo_t) );
  if(!P.p[P.np])
    {
      fprintf( stderr, _("%s Cannot store partition data: %s\n"),
	       progname, strerror(errno) );
      exit( EXIT_FAILURE );
    }
 
  /* set the inode and name of the new partition */
  P.p[P.np]->inode = inode;
  strncpy_t(P.p[P.np]->name, sizeof(P.p[P.np]->name), name, strlen(name));
  
  /* set the fstype and size of the new partition */  
  partition_get_fstype_from_blkid(name, &buf);
  strncpy_t(P.p[P.np]->fstype, sizeof(P.p[P.np]->fstype), buf, strlen(buf));
  free(buf);
  
  partition_get_size_from_lsblk(name, &buf);
  strncpy_t(P.p[P.np]->size, sizeof(P.p[P.np]->size), buf, strlen(buf));
  free(buf);
  		  
  partition_get_uuid_from_blkid(name, &buf);
  strncpy_t(P.p[P.np]->uuid, sizeof(P.p[P.np]->uuid), buf, strlen(buf));
  free(buf);
 
  P.p[P.np]->label[0] = P.p[P.np]->label[PM_LABEL_LEN] = '\0';
  P.p[P.np]->mnt[0] = P.p[P.np]->mnt[PM_MNT_LEN] = '\0';
  P.p[P.np]->nmounts = 0;
  

  if(P.p[P.np]->fstype && (!strcmp(P.p[P.np]->fstype, "swap") || !strncmp(P.p[P.np]->fstype, "--", 2)))
    return;
  
  if(P.p[P.np]->uuid) {

    char eol[1024]={0};
	char line[1024]={0};
	
    FILE *fp = NULL;
    fp = fopen("/etc/fstab", "r");
    if(!fp) {
      fprintf(stderr, "%s: fopen(%s): %s\n", progname, strerror(errno));
    }
  
    strncpy_t(line, sizeof(line), "UUID=", strlen("UUID="));
	strcat(line, P.p[P.np]->uuid);
    
    while(fgets(eol, sizeof(eol), fp)) {
	  eol[strcspn(eol, "\n")] = '\0';
      if(!strncmp(line, eol, strlen(line))) {
        fclose(fp);
        return;
      }
    }
    
    fclose(fp);
  }

  /* Tell ui we have a new partition */
  ui_create(P.p[P.np]);
  
  P.np ++;
  if(debug) partition_print();
  
  return;
}

/*------------------------------- Destructor --------------------------------*/
void partition_delete( const char *name )
{
  unsigned index;

  /* Find partition */
  for(index=0; index<P.np; index ++) if(!strcmp(P.p[index]->name, name)) break;
  if(index == P.np) return;

  /* Tell ui to remove partition from display */
  ui_delete(P.p[index]);

  /* free partition's data */
  free(P.p[index]);

  -- P.np;
  
  /* reshuffle the list */
  for( ; index < P.np; index ++) P.p[index] = P.p[index+1];
  
  /* reallocate memory for the new number of partitions */
  P.p = realloc( P.p, sizeof(partinfo_t *) * P.np );

  if( P.np && !P.p )
    {
      fprintf(stderr,_("%s Error freeing memory, returned by realloc(): %s\n"),
	       progname, strerror(errno) );
      exit( EXIT_FAILURE );
    }

  if(debug) partition_print();
  return;
}

/*------------------------------- Getters -----------------------------------*/
partlist_t partition_get_list(void)
{
  /* for use in mountpoints.c - dangerous, use with care ! */
  return P;
}

unsigned partition_by_name( const char *name )
{
  unsigned index;
  for(index=0; index<P.np; index++)
    {
      if( !strcmp(P.p[index]->name, name) ) break;
    }
  return index; /* If partition isn't found, return number of partitions */
}

unsigned partition_by_inode( ino_t inode )
{
  unsigned index;
  for(index=0; index<P.np; index++)
    {
      if( P.p[index]->inode == inode ) break;
    }
  return index; /* If partition isn't found, return number of partitions */
}

partinfo_t *get_partition( ino_t inode)
{
  int index;
  index = partition_by_inode(inode);
  return P.p[index];
}

/*--------------------------------- Setters ---------------------------------*/
void partition_set_label( ino_t inode, const char *label )
{
  unsigned index;
  int changed = 0;
  index = partition_by_inode(inode);
  if(index < P.np)
    {
      if( strcmp(P.p[index]->label, label) )
	{
	  strncpy(P.p[index]->label, label, PM_LABEL_LEN);
	  ui_update(P.p[index]);
	  changed ++;
	}
    }
  
  if(debug && changed) partition_print();
}

void partition_set_label_by_name( const char *name, const char *label )
{
  unsigned index;
  int changed = 0;
  index = partition_by_name(name);
  if(index < P.np)
    {
      if( strcmp(P.p[index]->label, label) )
	{
	  strncpy(P.p[index]->label, label, PM_LABEL_LEN);
	  ui_update(P.p[index]);
	  changed ++;
	}
    }
  
  if(debug && changed) partition_print();
}

void partition_set_label_by_blkid(const char *name)
{ 
  int rc;
  char *tmp = NULL;
  char cmd[256] = {0};
  char output[256];
  FILE *pf = NULL;
  
  char *target = (char*) malloc(sizeof(char) * NAME_MAX+1);
  if(!target) 
    {
      fprintf(stderr, _("%s: Memory allocation failure.\n"), progname);
      exit(EXIT_FAILURE);
    }

  rc = readlink(name, target, NAME_MAX+1);    
  if(rc < 0)
    {
      fprintf(stderr, "[WARN]%s: readlink('%s') %s\n", 
                    progname, name, strerror(errno));
      free(target);    
      return;
    }
	
  tmp = rindex(target, '/') + 1;
  strncpy_t(cmd, sizeof(cmd), "ubus call ering.usbmount get_label '{ \"device\": \"/dev/", strlen("ubus call ering.usbmount get_label '{ \"device\": \"/dev/"));  
  strcat(cmd, tmp);  
  strcat(cmd, "\" }'");
  
  pf = popen(cmd, "r");
  if(!pf) {
     fprintf(stderr, "%s: popen(%s): %s\n", progname, strerror(errno));
  }
  
  while(fgets(output, sizeof(output), pf)) {
	char *aux = NULL;
	char *lbl = NULL; 
    output[strcspn(output, "\n")] = '\0';
    if(strlen(output)<3) continue;
    aux = rindex( output, ':' ) + 3;  
    aux[strcspn(aux, "\"")] = '\0';
    lbl = rindex(aux, '|') + 1;
    // We cannot compare lbl with Label, because we have an issue 
    // related to the text format (spaces are shown as \x20). So,
    // we prefer to use lbl (gotten from blkid) instead of Label
    if(strcmp(lbl, "*")) {
      aux = strtok(aux, "|");
	  partition_set_label_by_name(aux, lbl);
      break;  
    }
  }
  
  pclose(pf);
  free(target);  
}

void partition_get_label_from_blkid(const char *name, char **buf)
{ 
  int rc;
  char *tmp = NULL;
  char cmd[256] = {0};
  char output[256];
  FILE *pf = NULL;
		
  *buf = (char*)malloc(sizeof(char) * PM_LABEL_LEN+1);
  if (!*buf) {
	fprintf(stderr, "Memory allocation failure\n");
	exit (EXIT_FAILURE);
  }
  
  char *target = (char*) malloc(sizeof(char) * NAME_MAX+1);
  if(!target) 
    {
      fprintf(stderr, _("%s: Memory allocation failure.\n"), progname);
      exit(EXIT_FAILURE);
    }

  rc = readlink(name, target, NAME_MAX+1);    
  if(rc < 0)
    {
      fprintf(stderr, "[WARN]%s: readlink('%s') %s\n", 
                    progname, name, strerror(errno));
      free(target);    
      return;
    }
	
  tmp = rindex(target, '/') + 1;
  strncpy_t(cmd, sizeof(cmd), "ubus call ering.usbmount get_label '{ \"device\": \"/dev/", strlen("ubus call ering.usbmount get_label '{ \"device\": \"/dev/"));  
  strcat(cmd, tmp);  
  strcat(cmd, "\" }'");
  
  pf = popen(cmd, "r");
  if(!pf) {
     fprintf(stderr, "%s: popen(%s): %s\n", progname, strerror(errno));
  }
  
  while(fgets(output, sizeof(output), pf)) {
	char *aux = NULL;
	char *lbl = NULL; 
    output[strcspn(output, "\n")] = '\0';
    if(strlen(output)<3) continue;
    aux = rindex( output, ':' ) + 3;  
    aux[strcspn(aux, "\"")] = '\0';
    lbl = rindex(aux, '|') + 1;
    if(strcmp(lbl, "*"))
       strcpy(*buf, lbl);
  }
  
  pclose(pf);
  free(target);  
}

void partition_get_uuid_from_blkid(const char *name, char **buf)
{ 
  int rc;
  char cmd[256] = {0};
  char output[256];
  FILE *pf = NULL;
		
  *buf = (char*)malloc(sizeof(char) * PM_LABEL_LEN+1);
  if (!*buf) {
	fprintf(stderr, "Memory allocation failure\n");
	exit (EXIT_FAILURE);
  }
  strncpy_t(cmd, sizeof(cmd), "ubus call ering.usbmount get_uuid '{ \"device\": \"/dev/", strlen("ubus call ering.usbmount get_uuid '{ \"device\": \"/dev/"));  
  strcat(cmd, name);  
  strcat(cmd, "\" }'");
  
  pf = popen(cmd, "r");
  if(!pf) {
     fprintf(stderr, "%s: popen(%s): %s\n", progname, strerror(errno));
  }
  
  while(fgets(output, sizeof(output), pf)) {
	char *aux = NULL;
	char *uuid = NULL; 
    output[strcspn(output, "\n")] = '\0';
    if(strlen(output)<3) continue;
    aux = rindex( output, ':' ) + 3;  
    aux[strcspn(aux, "\"")] = '\0';
    uuid = rindex(aux, '|') + 1;
    if(strcmp(uuid, "*"))
      strcpy(*buf, uuid);
  }
  
  pclose(pf);
}

void partition_get_fstype_from_blkid(const char *name, char **buf)
{ 
  int rc;
  char *tmp = NULL;
  char cmd[256] = {0};
  char output[256];
  FILE *pf = NULL;
		
  *buf = (char*)malloc(sizeof(char) * PM_FSTYPE_LEN+1);
  if (!*buf) {
	fprintf(stderr, "Memory allocation failure\n");
	exit (EXIT_FAILURE);
  }
	
  strncpy_t(cmd, sizeof(cmd), "ubus call ering.usbmount get_fstype '{ \"device\": \"/dev/", strlen("ubus call ering.usbmount get_fstype '{ \"device\": \"/dev/"));   
  strcat(cmd, name);
  strcat(cmd, "\" }'");

  pf = popen(cmd, "r");
  if(!pf) {
     fprintf(stderr, "%s: popen(%s): %s\n", progname, strerror(errno));
  }
  
  while(fgets(output, sizeof(output), pf)) {
	 char *aux = NULL;
	 char *fstype = NULL;
     output[strcspn(output, "\n")] = '\0';
     if(strlen(output)<3) continue;
     aux = rindex( output, ':' ) + 3;
     aux[strcspn(aux, "\"")] = '\0';
     fstype = rindex(aux, '|') + 1;
     if(strcmp(fstype, "*"))
        strcpy(*buf, fstype);
     else
        strcpy(*buf, "------");
  }
  
  pclose(pf);
}

void partition_get_size_from_blkid(const char *name, char **buf)
{ 
  int rc;
  char *tmp = NULL;
  char cmd[256] = {0};
  char output[256];
  FILE *pf = NULL;
		
  *buf = (char*)malloc(sizeof(char) * PM_SIZE_LEN+1);
  if (!*buf) {
	fprintf(stderr, "Memory allocation failure\n");
	exit (EXIT_FAILURE);
  }
	
  strncpy_t(cmd, sizeof(cmd), "ubus call ering.usbmount get_size '{ \"device\": \"/dev/", strlen("ubus call ering.usbmount get_size '{ \"device\": \"/dev/"));   
  strcat(cmd, name);
  strcat(cmd, "\" }'");

  pf = popen(cmd, "r");
  if(!pf) {
     fprintf(stderr, "%s: popen(%s): %s\n", progname, cmd, strerror(errno));
  }
  
  memset(buf, 0, sizeof(buf));
  while(fgets(output, sizeof(output), pf)) {
	 char *aux = NULL;
	 char *size = NULL;
     output[strcspn(output, "\n")] = '\0';
     if(strlen(output)<3) continue;
     aux = rindex( output, ':' ) + 3;
     aux[strcspn(aux, "\"")] = '\0';
     size = rindex(aux, '|') + 1;
     strcpy(*buf, size);
  }

  pclose(pf);
}

void partition_get_size_from_lsblk(const char *name, char **buf)
{ 
  int rc;
  char cmd[512] = {0};
  char output[32] = {0};
  FILE *pf = NULL;
		
  *buf = (char*)malloc(sizeof(char) * PM_SIZE_LEN+1);
  if (!*buf) {
	fprintf(stderr, "Memory allocation failure\n");
	exit (EXIT_FAILURE);
  }
	
  strncpy_t(cmd, sizeof(cmd), "/bin/lsblk -o NAME,SIZE | grep ", strlen("/bin/lsblk -o NAME,SIZE | grep "));  
  strcat(cmd, name);

  pf = popen(cmd, "r");
  if(!pf) {
     fprintf(stderr, "%s: popen(%s): %s\n", progname, strerror(errno));
  }
  
  while(fgets(output, sizeof(output), pf)) {
	 char *token = NULL; 
     output[strcspn(output, "\n")] = '\0';
     token = strtok(output, " ");
     token = strtok(NULL, " ");
     strcpy(*buf, token);
  }
  
  pclose(pf);
}

/*-------------------------------- Debugger ---------------------------------*/
void partition_print(void)
{
  int i;
  if(P.np)
    fprintf(stderr,"   Inode Nm Device-Name    Label          Fstype        "
	    " Mount-Point\n");
  for( i=0; i<P.np; i++ )
    {
      fprintf(stderr, "%8d %2d %-14s %-14s %-14s %-.20s\n",
	      P.p[i]->inode, P.p[i]->nmounts, P.p[i]->name, P.p[i]->label,
	      P.p[i]->fstype, P.p[i]->mnt);
    }
}

